from typing import TypedDict
import requests

class HttpApiConfig(TypedDict):
    agentlabs_url: str;
    project_id: str;
    secret: str;

class HttpApi:
    def __init__(self, agentlabs_url: str, project_id: str, secret: str) -> None:
        self.agentlabs_url = agentlabs_url
        self.project_id = project_id
        self.secret = secret
        self.client = requests.Session()
        self.client.params = {
            "projectId": self.project_id
        }
        self.client.headers.update({
            "x-agentlabs-sdk-secret": self.secret,
            "x-agentlabs-project-id": self.project_id,
        })

    def _make_url(self, path: str):
        return self.agentlabs_url + path



from datetime import datetime

class Logger:
    def __init__(self, name: str) -> None:
        self.name = name

    def _format_log(self, level: str, message: str) -> str:
        return f"[{self.name}] {datetime.now().strftime('%m/%d/%Y, %H:%M:%S')} {level} {message}"

    def _log(self, level: str, message: str) -> None:
        print(self._format_log(level, message))

    def info(self, message: str) -> None:
        self._log("INFO", message)

    def warn(self, message: str) -> None:
        self._log("WARN", message)

    def error(self, message: str) -> None:
        self._log("ERROR", message)

    def debug(self, message: str) -> None:
        self._log("DEBUG", message)



from enum import Enum
from typing import TypedDict

class _ChatMessage(TypedDict):
    text: str
    conversationId: str
    messageId: str
    agentId: str
    memberId: str

class MessageFormat(Enum):
    PLAIN_TEXT = "PLAIN_TEXT"
    MARKDOWN = "MARKDOWN"

class IncomingChatMessage:
    """Represents a chat message received from the AgentLabs server.
    """
    def __init__(self, message: _ChatMessage):
        self.text = message["text"]
        self.conversation_id = message["conversationId"]
        self.message_id = message["messageId"]
        self.member_id = message["memberId"]



import os
from typing import Any, Callable, Dict

from .agent import Agent
from .chat import IncomingChatMessage

from ._internals.http import HttpApi
from ._internals.logger import Logger
from ._internals.realtime import RealtimeClient

class Project:
    """Represents a project on the AgentLabs server.
    This class is used to instantiate a backend connection for
    the configured project.
    """

    _client_logger = Logger(name="Client")
    _server_logger = Logger(name="Server")

    def _log_message(self, payload: Dict[str, Any]):
        message = payload.get('message')
        if not message is None:
            self._server_logger.info(message)

    def _handle_hearbeat(self, payload: Dict[str, Any]):
        if self.is_debug_enabled:
            self._client_logger.debug("Server heartbeat acknowledged.")

        return {
                'ok': True
        }

    def __init__(self, agentlabs_url: str, project_id: str, secret: str) -> None:
        self.is_debug_enabled = bool(os.environ.get('DEBUG', False))
        self.agentlabs_url = agentlabs_url
        self.project_id = project_id
        self._secret = secret
        self._http = HttpApi(
            project_id=project_id,
            agentlabs_url=agentlabs_url,
            secret=secret
        )
        self._realtime = RealtimeClient(project_id=project_id, secret=secret, url=agentlabs_url)
        self._realtime.on('message', self._log_message)
        self._realtime.on('heartbeat', self._handle_hearbeat)
   
    def on_chat_message(self, fn: Callable[[IncomingChatMessage], None]):
        """Defines a handler for when a new chat message is received.
        It will be called each time a member of your project sends a new message.
        """
        def wrapper(payload: Any):
            chat_message = IncomingChatMessage(message=payload['data'])
            fn(chat_message)

        self._realtime.on('chat-message', wrapper)

    def connect(self):
        """Connects the project to the AgentLabs server.
        Does not block the main thread by itself, use wait() if this is desired.
        May raise an exception if the connection fails.

        Note that as of now, only one connection per project is permitted.
        This will be changed very soon.
        """
        self._client_logger.info("Connecting to AgentLabs...")
        self._realtime.connect()

    def wait(self):
        """Blocks the main thread until the agent is disconnected
        Useful if you have only one agent and want to keep the program running
        without having to bother with your own loop.
        """
        self._realtime.wait()

    def disconnect(self):
        """Interrupt the current project backend connection.
        If this is the only backend connection used for the project, the project
        will be considered offline and unusable by members.

        Support for multiple connections per project is not yet implemented
        and will be added soon.
        """
        self._realtime.disconnect()

    def agent(self, id: str) -> Agent:
        """Embodies an agent defined for the project.
        This agent can be used to send and stream messages.
        """
        return Agent(
            realtime=self._realtime,
            id=id,
        )



from uuid import uuid4

from .chat import MessageFormat

from ._internals.realtime import RealtimeClient

class AgentStream:
    """Creates a stream to send a message in multiple parts.
        This is done by making successive calls to write() and then calling end() when done.
    """
    is_ended: bool = False

    def __init__(self, realtime: RealtimeClient, agent_id: str, conversation_id: str, format: MessageFormat):
        self.conversation_id = conversation_id
        self.message_id = str(uuid4())
        self.format = format
        self._realtime = realtime
        self.agent_id = agent_id
    
    def write(self, token: str):
        """Writes a token to the stream. This can be used to send a message in multiple parts.
        Writing to a stream on which end() has been called will raise an exception.
        """
        if self.is_ended:
            raise Exception("Cannot write to a stream that has already been ended.")

        self._realtime.emit('stream-chat-message-token', {
            "conversationId": self.conversation_id,
            "messageId": self.message_id,
            "text": token,
            "attachments": [],
            "format": self.format.value,
            "agentId": self.agent_id
        })
    
    def end(self):
        """
        Ends the stream. After a stream is ended it cannot be written to anymore, and
        doing so will raise an exception.
        """
        self.is_ended = True
        self._realtime.emit('stream-chat-message-end', {
            "conversationId": self.conversation_id,
            "messageId": self.message_id,
            "agentId": self.agent_id
        })

class Agent:
    def __init__(self, id: str, realtime: RealtimeClient) -> None:
        self._realtime = realtime
        self.id = id

    def send(self, text: str, conversation_id: str, format: MessageFormat = MessageFormat.PLAIN_TEXT):
        """Sends a message to a conversation.
        Such message will be communicated instantly to the conversation.
        """
        self._realtime.emit('chat-message', {
            "conversationId": conversation_id,
            "text": text,
            "agentId": self.id,
            "source": "AGENT",
            "format": format.value
        })

    def create_stream(self, conversation_id: str, format: MessageFormat) -> AgentStream:
        """Creates a stream to send a message in multiple parts.
        This is done by making successive calls to write() and then calling end() when done.
        """
        return AgentStream(self._realtime, self.id, conversation_id, format)
